import hashlib
import pathlib
import pkg_resources
import signal
import sys
import traceback
import warnings

import dclab
from dclab.cli import repack
from dclab.cli import common

import h5py
import hdf5plugin
import imageio
import imageio_ffmpeg
import nptdms
import numpy
from PyQt5 import uic, QtCore, QtWidgets

from . import dlg_icheck
from .dlg_icheck import IntegrityCheckDialog
from . import history
from . import message_box
from . import meta_tool
from . import preferences
from . import update
from .wait_cursor import show_wait_cursor, ShowWaitCursor
from ._version import version


class DCKit(QtWidgets.QMainWindow):
    def __init__(self):
        super(DCKit, self).__init__()
        path_ui = pkg_resources.resource_filename("dckit", "main.ui")
        uic.loadUi(path_ui, self)
        self.setWindowTitle(f"DCKit {version}")
        # update check
        self._update_thread = None
        self._update_worker = None
        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        QtCore.QCoreApplication.setOrganizationName("DC-Analysis")
        QtCore.QCoreApplication.setOrganizationDomain("dc.readthedocs.io")
        QtCore.QCoreApplication.setApplicationName("DCKit")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
        #: DCKit settings
        self.settings = QtCore.QSettings()
        self.settings.setIniCodec("utf-8")
        # reload temporary features
        preferences.register_temporary_features()
        # Disable native menubar (e.g. on Mac)
        self.menubar.setNativeMenuBar(False)
        # signals
        self.pushButton_integrity.clicked.connect(self.on_task_integrity_all)
        self.pushButton_compress.clicked.connect(self.on_task_compress)
        self.pushButton_metadata.clicked.connect(self.on_task_metadata)
        self.pushButton_split.clicked.connect(self.on_task_split)
        self.pushButton_tdms2rtdc.clicked.connect(self.on_task_tdms2rtdc)
        self.pushButton_join.clicked.connect(self.on_task_join)
        self.tableWidget.itemChanged.connect(self.on_table_text_changed)
        self.checkBox_repack.clicked.connect(self.on_repack)
        # File menu
        self.action_add.triggered.connect(self.on_action_add_measurements)
        self.action_add_folder.triggered.connect(self.on_action_add_folder)
        self.action_clear.triggered.connect(self.on_action_clear_measurements)
        self.action_quit.triggered.connect(self.on_action_quit)
        # Edit menu
        self.action_preferences.triggered.connect(self.on_action_preferences)
        # Help menu
        self.actionSoftware.triggered.connect(self.on_action_software)
        self.actionAbout.triggered.connect(self.on_action_about)
        #: contains all imported paths (index is DCKit-id)
        self.pathlist = []
        # contains all integrity buttons (keys are DCKit-ids)
        self.integrity_buttons = {}
        # if "--version" was specified, print the version and exit
        if "--version" in sys.argv:
            print(version)
            QtWidgets.QApplication.processEvents(
                QtCore.QEventLoop.AllEvents, 300)
            sys.exit(0)
        # check update after version printing (QThread: Destroyed error)
        do_update = int(self.settings.value("check for updates", 1))
        self.on_action_check_update(do_update)
        self.show()
        self.raise_()
        self.activateWindow()
        self.setWindowState(QtCore.Qt.WindowState.WindowActive)

    @show_wait_cursor
    def append_paths(self, pathlist):
        """Append selected paths to table"""
        datas = []
        # get meta data for all paths
        for path in pathlist:
            try:  # avoid any errors
                info = {"DCKit-id": (0, len(self.pathlist)),
                        "integrity": (1, None),
                        "path": (2, path),
                        "sample": (3, meta_tool.get_sample_name(path)),
                        "run index": (4, meta_tool.get_run_index(path)),
                        "event count": (5, meta_tool.get_event_count(path)),
                        "flow rate": (6, meta_tool.get_flow_rate(path)),
                        "chip region": (None, meta_tool.get_chip_region(path)),
                        }
            except BaseException:
                warnings.warn("Could not append dataset {} ".format(path)
                              + "(traceback follows)!\n"
                              + "{}".format(traceback.format_exc()))
                # stop doing anything
                continue
            self.pathlist.append(pathlib.Path(path))
            datas.append(info)
        # populate table widget
        for info in datas:
            row = self.tableWidget.rowCount()
            self.tableWidget.insertRow(row)
            for key in info:
                col, val = info[key]
                if col is None:
                    continue
                if key == "integrity":
                    # button
                    btn = QtWidgets.QToolButton(self)
                    btn.setText("run check")
                    self.tableWidget.setCellWidget(row, col, btn)
                    btn.clicked.connect(self.on_integrity_check)
                    self.integrity_buttons[info["DCKit-id"][1]] = btn
                else:
                    # text
                    item = QtWidgets.QTableWidgetItem("{}".format(val))
                    if key == "sample":
                        # allow editing sample name
                        item.setFlags(QtCore.Qt.ItemIsEnabled
                                      | QtCore.Qt.ItemIsEditable)
                    elif key == "path":
                        item.setText(pathlib.Path(val).name)
                        item.setToolTip(str(val))
                        item.setFlags(QtCore.Qt.ItemIsEnabled)
                    elif key == "flow rate":
                        if info["chip region"][1] == "channel":
                            item.setText("{:.5f}".format(val))
                        else:
                            item.setText("reservoir")
                        item.setFlags(QtCore.Qt.ItemIsEnabled)
                    else:
                        item.setFlags(QtCore.Qt.ItemIsEnabled)
                    self.tableWidget.setItem(row, col, item)
        if datas:
            # set header widths
            self.tableWidget.setColumnWidth(info["DCKit-id"][0], 10)
            self.tableWidget.setColumnWidth(info["integrity"][0], 100)
            self.tableWidget.setColumnWidth(info["path"][0], 180)
            self.tableWidget.setColumnWidth(info["run index"][0], 80)
            self.tableWidget.setColumnWidth(info["flow rate"][0], 100)
            self.tableWidget.setColumnWidth(info["event count"][0], 80)
            self.tableWidget.setColumnWidth(info["sample"][0], 300)

    def dragEnterEvent(self, e):
        """Whether files are accepted"""
        if e.mimeData().hasUrls():
            e.accept()
        else:
            e.ignore()

    def dropEvent(self, e):
        """Add dropped files to view"""
        urls = e.mimeData().urls()
        pathlist = []
        for ff in urls:
            pp = pathlib.Path(ff.toLocalFile())
            if pp.is_dir():
                pathlist += meta_tool.find_data(pp)
            elif pp.suffix in [".rtdc", ".tdms"]:
                pathlist.append(pp)
        self.append_paths(pathlist)

    def get_metadata(self, row):
        path = self.get_path(row)
        metadata = IntegrityCheckDialog.metadata_from_path(path)
        # update sample name
        newname = self.tableWidget.item(row, 3).text()
        if "experiment" not in metadata:
            metadata["experiment"] = {}
        metadata["experiment"]["sample"] = newname
        return metadata

    def get_path(self, row):
        """Return dataset path from a given row

        This is necessary, because the user can sort columns
        """
        path_index = int(self.tableWidget.item(row, 0).text())
        path = self.pathlist[path_index]
        return path

    @QtCore.pyqtSlot()
    def on_action_add_folder(self):
        """Search folder for RT-DC data and add to table"""
        # show a dialog for selecting folder
        path = QtWidgets.QFileDialog.getExistingDirectory(self)
        if not path:
            return
        # find RT-DC data
        pathlist = meta_tool.find_data(path)
        if not pathlist:
            raise ValueError("No RT-DC data found in {}!".format(path))
        # add to list
        self.append_paths(pathlist)

    @QtCore.pyqtSlot()
    def on_action_add_measurements(self):
        """Select .tdms and .rtdc files and add to table"""
        # show a dialog for adding multiple single files (.tdms and .rtdc)
        pathlist, _ = QtWidgets.QFileDialog.getOpenFileNames(
            self,
            'Select RT-DC data',
            '',
            'RT-DC data (*.tdms *.rtdc)')
        if pathlist:
            # add to list
            self.append_paths(pathlist)

    @QtCore.pyqtSlot()
    def on_action_about(self):
        about_text = "DCKit is a tool for managing RT-DC data.\n\n" \
            + "Author: Paul Müller\n" \
            + "Code: https://github.com/DC-analysis/DCKit\n"
        QtWidgets.QMessageBox.about(self, f"DCKit {version}", about_text)

    @QtCore.pyqtSlot(bool)
    def on_action_check_update(self, b):
        self.settings.setValue("check for updates", int(b))
        if b and self._update_thread is None:
            self._update_thread = QtCore.QThread()
            self._update_worker = update.UpdateWorker()
            self._update_worker.moveToThread(self._update_thread)
            self._update_worker.finished.connect(self._update_thread.quit)
            self._update_worker.data_ready.connect(
                self.on_action_check_update_finished)
            self._update_thread.start()

            ghrepo = "DC-analysis/DCKit"

            QtCore.QMetaObject.invokeMethod(self._update_worker,
                                            'processUpdate',
                                            QtCore.Qt.QueuedConnection,
                                            QtCore.Q_ARG(str, version),
                                            QtCore.Q_ARG(str, ghrepo),
                                            )

    @QtCore.pyqtSlot(dict)
    def on_action_check_update_finished(self, mdict):
        # cleanup
        self._update_thread.quit()
        self._update_thread.wait()
        self._update_worker = None
        self._update_thread = None
        # display message box
        ver = mdict["version"]
        web = mdict["releases url"]
        dlb = mdict["binary url"]
        msg = QtWidgets.QMessageBox()
        msg.setWindowTitle("DCKit {} available!".format(ver))
        msg.setTextFormat(QtCore.Qt.RichText)
        text = "You can install DCKit {} ".format(ver)
        if dlb is not None:
            text += 'from a <a href="{}">direct download</a>. '.format(dlb)
        else:
            text += 'by running `pip install --upgrade dckit`. '
        text += 'Visit the <a href="{}">official release page</a>!'.format(web)
        msg.setText(text)
        msg.exec_()

    @QtCore.pyqtSlot()
    def on_action_clear_measurements(self):
        """Clear the table"""
        for _ in range(len(self.pathlist)):
            self.tableWidget.removeRow(0)
        self.pathlist.clear()
        self.integrity_buttons.clear()
        # clear lru_cache
        meta_tool.get_rtdc_meta.cache_clear()
        dlg_icheck.check_dataset.cache_clear()

    @QtCore.pyqtSlot()
    def on_action_preferences(self):
        """Show the preferences dialog"""
        dlg = preferences.Preferences(self)
        dlg.setWindowTitle("DCKit Preferences")
        dlg.exec()

    @QtCore.pyqtSlot()
    def on_action_software(self):
        libs = [dclab,
                h5py,
                imageio,
                imageio_ffmpeg,
                nptdms,
                numpy,
                ]
        sw_text = f"DCKit {version}\n\n"
        sw_text += f"Python {sys.version}\n\n"
        sw_text += "Modules:\n"
        sw_text += f"- hdf5plugin {hdf5plugin.version}\n"
        for lib in libs:
            sw_text += f"- {lib.__name__} {lib.__version__}\n"
        sw_text += f"- PyQt5 {QtCore.QT_VERSION_STR}\n"
        if hasattr(sys, 'frozen'):
            sw_text += "\nThis executable has been created using PyInstaller."
        QtWidgets.QMessageBox.information(self, "Software", sw_text)

    @QtCore.pyqtSlot()
    def on_action_quit(self):
        """Determine what happens when the user wants to quit"""
        QtCore.QCoreApplication.quit()

    @QtCore.pyqtSlot()
    def on_integrity_check(self, button=None):
        if button is None:
            button = self.sender()
            skip_ui = False
        else:
            skip_ui = True
        # find DCKit-id
        for did in self.integrity_buttons:
            if button is self.integrity_buttons[did]:
                break
        else:
            raise ValueError("Could not find button {}".format(button))
        # get path
        path = self.pathlist[did]
        with ShowWaitCursor():
            dlg = IntegrityCheckDialog(self, path)
        if skip_ui:
            dlg.done(True)
        else:
            dlg.exec_()
        button.setText(dlg.state)
        colors = {"failed": "#A50000",
                  "tolerable": "#7A6500",
                  "passed": "#007A04"}
        button.setStyleSheet("color: {}".format(colors[dlg.state]))

    @QtCore.pyqtSlot()
    def on_repack(self):
        """The checkbox is clicked (no repacking is performed)"""
        if self.checkBox_repack.isChecked():
            # ask the user whether he knows what he is doing
            dlg = QtWidgets.QDialog()
            path_ui = pkg_resources.resource_filename("dckit", "dlg_repack.ui")
            uic.loadUi(path_ui, dlg)
            ret = dlg.exec_()
            if ret == QtWidgets.QDialog.Rejected:
                self.checkBox_repack.setChecked(False)
                self.pushButton_metadata.setEnabled(True)
            else:
                self.pushButton_metadata.setEnabled(False)
        else:
            self.pushButton_metadata.setEnabled(True)

    @QtCore.pyqtSlot()
    def on_table_text_changed(self):
        """Reset sample name if set to empty string"""
        curit = self.tableWidget.currentItem()
        if curit is not None and curit.text() == "":
            row = self.tableWidget.currentRow()
            path = self.get_path(row)
            sample = meta_tool.get_sample_name(path)
            self.tableWidget.item(row, 3).setText(sample)

    @QtCore.pyqtSlot()
    def on_task_compress(self):
        """Compress .rtdc data losslessly"""
        # Open the target directory
        pout = QtWidgets.QFileDialog.getExistingDirectory(self)
        details = []
        invalid = []
        paths_compressed = []
        if pout:
            with ShowWaitCursor():
                pout = pathlib.Path(pout)
                for row in range(self.tableWidget.rowCount()):
                    path = self.get_path(row)
                    metadata = self.get_metadata(row)
                    name = metadata["experiment"]["sample"]
                    prtdc = pout / get_rtdc_output_name(origin_path=path,
                                                        sample_name=name)
                    if path.suffix == ".rtdc":
                        task_dict = {
                            "name": "compress HDF5 data",
                        }
                        dclab.cli.compress(path_in=path, path_out=prtdc)
                        append_execution_log(prtdc, task_dict)
                        task_dict_meta = self.write_metadata(prtdc, metadata)
                        if task_dict_meta:
                            append_execution_log(prtdc, task_dict_meta)
                        # write any warnings to separate log files
                        extract_warning_logs(prtdc)
                        # update list for UI
                        details.append("{} -> {}".format(path, prtdc))
                        paths_compressed.append(prtdc)
                        # repack if checked
                        # (we still did the compression in case dclab needed
                        # to fix things)
                        self.repack(prtdc)
                    else:
                        # do not do anything with .rtdc files
                        invalid.append(path)
        else:
            return

        if invalid:
            message_box.ignored(
                message=f"{len(invalid)} files were ignored (no .rtdc data).",
                details="Affected files are:\n"
                        + "\n\n".join([str(p) for p in invalid]))

        # finally, show the feedback dialog
        if len(details):
            message_box.success(
                message=f"Successfully compressed {len(details)} .rtdc files!",
                details="\n\n".join(details))
        elif not invalid:
            message_box.nothing_todo()
        return paths_compressed, invalid

    @show_wait_cursor
    @QtCore.pyqtSlot()
    def on_task_integrity_all(self):
        for did in self.integrity_buttons:
            btn = self.integrity_buttons[did]
            self.on_integrity_check(button=btn)

    @QtCore.pyqtSlot()
    def on_task_join(self):
        """Join multiple RT-DC measurements"""
        # show a dialog with sample name
        dlg = QtWidgets.QDialog()
        path_ui = pkg_resources.resource_filename("dckit", "dlg_join.ui")
        uic.loadUi(path_ui, dlg)
        dlg.lineEdit.setText(self.get_metadata(0)["experiment"]["sample"])
        ret = dlg.exec_()
        if ret == QtWidgets.QDialog.Accepted:
            sample = dlg.lineEdit.text()
            metadata = {"experiment": {"run index": 1,
                                       "sample": sample}}
            po, _ = QtWidgets.QFileDialog.getSaveFileName(
                self, 'Output path', '', 'RT-DC files (*.rtdc)')
            if po:
                po = pathlib.Path(po)
                if not po.suffix == ".rtdc":
                    po = po.parent / (po.name + ".rtdc")
                pi = []
                for row in range(self.tableWidget.rowCount()):
                    pi.append(self.get_path(row))
                # finally, show the feedback dialog
                msg = QtWidgets.QMessageBox()
                if pi:
                    with ShowWaitCursor():
                        dclab.cli.join(path_out=po,
                                       paths_in=pi,
                                       metadata=metadata)
                    # write any warnings to separate log files
                    extract_warning_logs(po)
                    # repack if checked
                    self.repack(po)
                    # display message
                    msg.setIcon(QtWidgets.QMessageBox.Information)
                    msg.setText(f"Successfully joined {len(pi)} datasets!")
                    msg.setWindowTitle("Success")
                    msg.setDetailedText("\n".join([str(pp) for pp in pi]))
                else:
                    msg.setIcon(QtWidgets.QMessageBox.Warning)
                    msg.setText("Nothing to do!")
                    msg.setWindowTitle("Warning")
                msg.exec_()

    @QtCore.pyqtSlot()
    def on_task_metadata(self):
        """Update the metadata including the sample names of the datasets"""
        invalid = []
        details = []
        with ShowWaitCursor():
            for row in range(self.tableWidget.rowCount()):
                path = self.get_path(row)
                # check whether we are allowed to do this
                if path.suffix == ".tdms":
                    # not supported for tdms files
                    invalid.append(path)
                else:
                    metadata = self.get_metadata(row)
                    task_dict_meta = self.write_metadata(path, metadata)
                    if task_dict_meta:
                        append_execution_log(path, task_dict_meta)
                        # update list for UI
                        details.append("{}: update metadata".format(path))
                        # remove item from check cache
                        if path in IntegrityCheckDialog.user_metadata:
                            IntegrityCheckDialog.user_metadata.pop(path)
                    dlg_icheck.check_dataset.cache_clear()
        if invalid:
            # show message regarding .tdms data
            message_box.ignored(
                message=f"{len(invalid)} files were ignored (no .rtdc data).",
                info="Changing the metadata for .tdms files is "
                     + "not supported! Please convert the files to "
                     + "the .rtdc file format.",
                details="Affected files are:\n"
                        + "\n\n".join([str(p) for p in invalid]))

        # finally, show the feedback dialog
        if len(details):
            message_box.success(
                message=f"Updated the metadata of {len(details)} files!",
                details="\n\n".join(details))
        elif not invalid:
            message_box.nothing_todo(
                message="Nothing to update! You did not edit any metadata in "
                        + "the 'Sample' column or via the integrity check or "
                        + "you are attempting to update .tdms data (which "
                        + "is not supported - use 'Convert .tdms to .rtdc' "
                        + "instead).")

    @QtCore.pyqtSlot()
    def on_task_split(self):
        split_events, ok_pressed = QtWidgets.QInputDialog.getInt(
            self, "Events per output file", "Limit events to:",
            10000, 0, 1000000, 5000)
        details = []
        errors = []
        paths_split = []

        if ok_pressed:
            pout = QtWidgets.QFileDialog.getExistingDirectory(self)
            if pout:
                with ShowWaitCursor():
                    pout = pathlib.Path(pout)
                    task_dict = {
                        "name": "split every {} events".format(split_events),
                    }
                    for row in range(self.tableWidget.rowCount()):
                        path = self.get_path(row)
                        try:
                            psplit = dclab.cli.split(
                                path_in=path,
                                path_out=pout,
                                split_events=split_events,
                                skip_initial_empty_image=True,
                                skip_final_empty_image=True,
                                ret_out_paths=True,
                                verbose=False)
                        except BaseException:
                            errors.append([path, traceback.format_exc()])
                            # remove erroneous files
                            for pp in pout.glob(path.stem+"_*"):
                                pp.unlink()
                        else:
                            for pp in psplit:
                                append_execution_log(pp, task_dict)
                                # write any warnings to separate log files
                                extract_warning_logs(pp)
                                # update list for UI
                                details.append("created {}".format(pp))
                                # repack if checked
                                self.repack(pp)
                            paths_split.append(path)
        if errors:
            # Show an error dialog for the files that could not be converted
            message_box.error(
                message="Errors encountered while splitting!",
                details="Affected files are:\n\n"
                        + "\n\n".join(["{}:\n{}".format(*e) for e in errors]))

        # finally, show the feedback dialog
        if len(details):
            message_box.success(
                message=f"Splitting into {len(details)} files completed.",
                details="\n\n".join(details))
        elif not errors:
            message_box.nothing_todo()
        return paths_split, errors

    @QtCore.pyqtSlot()
    def on_task_tdms2rtdc(self):
        """Convert .tdms files to .rtdc files"""
        pout = QtWidgets.QFileDialog.getExistingDirectory(self)
        details = []
        errors = []
        invalid = []
        paths_converted = []
        if pout:
            pout = pathlib.Path(pout)
            with ShowWaitCursor():
                for row in range(self.tableWidget.rowCount()):
                    path = self.get_path(row)
                    metadata = self.get_metadata(row)
                    name = metadata["experiment"]["sample"]
                    prtdc = pout / get_rtdc_output_name(origin_path=path,
                                                        sample_name=name)
                    if path.suffix == ".tdms":
                        task_dict = {
                            "name": "convert .tdms to .rtdc",
                        }
                        try:
                            dclab.cli.tdms2rtdc(path_tdms=path,
                                                path_rtdc=prtdc,
                                                compute_features=False,
                                                skip_initial_empty_image=True,
                                                skip_final_empty_image=True,
                                                verbose=False)
                        except BaseException:
                            errors.append([path, traceback.format_exc()])
                            if prtdc.exists():
                                prtdc.unlink()
                        else:
                            append_execution_log(prtdc, task_dict)
                            task_dict_meta = self.write_metadata(prtdc,
                                                                 metadata)
                            if task_dict_meta:
                                append_execution_log(prtdc, task_dict_meta)
                            # write any warnings to separate log files
                            extract_warning_logs(prtdc)
                            # update list for UI
                            details.append("{} -> {}".format(path, prtdc))
                            paths_converted.append(prtdc)
                            # repack if checked
                            self.repack(prtdc)
                    else:
                        # do not do anything with .rtdc files
                        invalid.append(path)
        if invalid:
            # Show an error dialog for the tdms files
            message_box.ignored(
                message="Only .tdms files supported as input, got "
                        + f"{len(invalid)} other files!",
                details="Affected files are:\n\n"
                        + "\n\n".join([str(p) for p in invalid]))

        if errors:
            message_box.error(
                message=f"{len(errors)} files could not be converted!",
                details="Affected files are:\n\n"
                        + "\n\n".join(["{}:\n{}".format(*e) for e in errors]))

        # finally, show the feedback dialog
        if len(details):
            message_box.success(
                message=f"Converted {len(details)} files to .rtdc.",
                details="\n\n".join(details))
        elif not (errors or invalid):
            message_box.nothing_todo()
        return paths_converted, invalid, errors

    @show_wait_cursor
    def repack(self, path):
        """repack and strip logs if the checkbox is checked"""
        if not self.checkBox_repack.isChecked():
            return

        path = pathlib.Path(path)
        # rename the original file
        path_orig = path.with_name(path.stem + "_original.rtdc")
        path.rename(path_orig)
        try:
            repack(path_in=path_orig, path_out=path, strip_logs=True)
        except BaseException:
            path_orig.unlink()
            if path.exists():
                path.unlink()
            raise
        else:
            path_orig.unlink()

    @show_wait_cursor
    def write_metadata(self, path, metadata):
        """Write metadata to an HDF5 file

        Returns
        -------
        `task_dict` if anything changed or None
        """
        # entry for the log
        task_dict = {
            "name": "update metadata",
            "old": {},
            "new": {},
        }
        # update in-place
        with h5py.File(path, "a") as h5:
            for sec in metadata:
                for key in metadata[sec]:
                    h5key = "{}:{}".format(sec, key)
                    value = metadata[sec][key]
                    value_old = h5.attrs.get(h5key, None)
                    if isinstance(value_old, bytes):
                        value_old = value_old.decode("utf-8")
                    if isinstance(value, bytes):
                        value = value.decode("utf-8")
                    if value != value_old:
                        task_dict["new"][h5key] = value
                        task_dict["old"][h5key] = value_old
                        if isinstance(value, str):  # (after task_dict)
                            value = numpy.string_(value.encode("utf-8"))
                        h5.attrs[h5key] = value
        if task_dict["new"]:
            return task_dict
        else:
            return None


def append_execution_log(path, task_dict):
    info = common.get_job_info()
    info["libraries"]["dckit"] = version
    info["task"] = task_dict
    history.append_history(path, info)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.

    :param `etype`: the exception type (`SyntaxError`,
        `ZeroDivisionError`, etc...);
    :type `etype`: `Exception`
    :param string `value`: the exception error message;
    :param string `trace`: the traceback header, if any (otherwise, it
        prints the standard Python header: ``Traceback (most recent
        call last)``.
    """
    vinfo = f"Unhandled exception in DCKit version {version}:\n"
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec_()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Clipboard)
        cb.setText(exception)


def extract_warning_logs(path):
    path = pathlib.Path(path)
    logs = meta_tool.get_rtdc_logs(path)
    for lname in logs:
        if lname.count("warnings"):
            plog = path.with_name(path.stem + "_" + lname + ".log")
            plog.write_text("\r\n".join(logs[lname]))


def get_rtdc_output_name(origin_path, sample_name):

    if meta_tool.get_chip_region(origin_path) == "channel":
        fl_or_res = "{:.4f}uls".format(meta_tool.get_flow_rate(origin_path))
    else:
        fl_or_res = "reservoir"

    name = "{}_M{}_{}_{}_{}.rtdc".format(
        meta_tool.get_date(origin_path),
        meta_tool.get_run_index(origin_path),
        fl_or_res,
        sample_name,
        sha256(origin_path)[:8])
    return get_valid_filename(name)


def get_valid_filename(value):
    """
    Return the given string converted to a string that can be used
    for a clean filename.
    """
    ret = ""

    valid = "abcdefghijklmnopqrstuvwxyz" \
            + "ABCDEFGHIJKLMNOPQRSTUVWXYZ" \
            + "0123456789" \
            + "._-()"
    replace = {
        " ": "_",
        "[": "(",
        "]": ")",
        "µ": "u",
    }

    for ch in value:
        if ch in valid:
            ret += ch
        elif ch in replace:
            ret += replace[ch]
        else:
            ret += "?"

    ret = ret.strip(".")
    return ret


def sha256(path):
    return dclab.util.hashfile(path, constructor=hashlib.sha256)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



import json
import numbers
import pathlib

import dclab
import h5py


def append_history(path, hdict):
    """Append DCKit history to an .rtdc file

    Parameters
    ----------
    path: str or pathlib.Path
        Path to .rtdc file
    hdict: dict
        History element (not the full history)
    """
    path = pathlib.Path(path)
    hlist = read_history(path)
    hlist.append(hdict)
    write_history(path, hlist)


def default_json_converter(obj):
    """is a function that should return a serializable version
    of obj or raise TypeError.
    """
    if isinstance(obj, numbers.Integral):
        return int(obj)
    elif isinstance(obj, bytes):
        return obj.decode("utf-8")
    else:
        raise TypeError(
            "Object of type '{}' is not JSON serializable".format(type(obj)))


def read_history(path):
    """Read DCKit history from an .rtdc file

    Parameters
    ----------
    path: str or pathlib.Path
        Path to .rtdc file

    Returns
    -------
    hlist: list of dicts
        Full history (containing history elements)
    """
    path = pathlib.Path(path)
    with dclab.new_dataset(path) as ds:
        if "dckit-history" in ds.logs:
            hdump = ds.logs["dckit-history"]
            hlist = json.loads("\n".join(hdump))
        else:
            hlist = []
    return hlist


def write_history(path, hlist):
    """Write DCKit history to .rtdc file

    Parameters
    ----------
    path: str or pathlib.Path
        Path to .rtdc file
    hlist: list of dicts
        Full history (containing history elements)
    """
    path = pathlib.Path(path)
    # remove previous log
    with h5py.File(path, "a") as h5:
        if "dckit-history" in h5["logs"]:
            del h5["logs"]["dckit-history"]
    # dump json log
    hlog = json.dumps(hlist,
                      sort_keys=True,
                      indent=2,
                      default=default_json_converter,
                      ).split("\n")
    # write dump as log
    with dclab.RTDCWriter(path) as hw:
        hw.store_log("dckit-history", hlog)



import copy
import functools
import pathlib
import warnings

import dclab
from dclab.rtdc_dataset import config as rt_config
from dclab.rtdc_dataset import fmt_tdms
import h5py
import nptdms


class MetadataEditedWarning(UserWarning):
    pass


def find_data(path):
    """Find tdms and rtdc data files in a directory"""
    path = pathlib.Path(path)

    def sort_path(path):
        """Sorting key for intuitive file sorting

        This sorts a list of RT-DC files according to measurement number,
        e.g. (M2_*.tdms is not sorted after M11_*.tdms):

        /path/to/M1_*.tdms
        /path/to/M2_*.tdms
        /path/to/M10_*.tdms
        /path/to/M11_*.tdms

        Note that the measurement number of .rtdc files is extracted from
        the hdf5 metadata and not from the file name.
        """
        try:
            # try to get measurement number as an integer
            idx = get_run_index(path)
        except BaseException:
            # just use the given path
            name = path.name
        else:
            # assign new "name" for sorting
            name = "{:09d}_{}".format(idx, path.name)
        return path.with_name(name)

    tdmsfiles = fmt_tdms.get_tdms_files(path)
    tdmsfiles = sorted(tdmsfiles, key=sort_path)
    rtdcfiles = [r for r in path.rglob("*.rtdc") if r.is_file()]
    rtdcfiles = sorted(rtdcfiles, key=sort_path)

    files = [pathlib.Path(ff) for ff in rtdcfiles + tdmsfiles]
    return files


def get_chip_region(path):
    fname = pathlib.Path(path).resolve()
    ext = fname.suffix
    if ext == ".rtdc":
        with h5py.File(fname, mode="r") as h5:
            chip_region = h5.attrs["setup:chip region"]
    elif ext == ".tdms":
        name = fname.name
        path = fname.parent
        mx = name.split("_")[0]
        para = path / (mx + "_para.ini")
        if para.exists():
            camcfg = rt_config.load_from_file(para)
            chip_region = camcfg["general"]["region"].lower()
        else:
            config = get_rtdc_config(path)
            chip_region = config["setup"]["chip region"]
    return chip_region


def get_date(path):
    config = get_rtdc_config(path)
    return config["experiment"]["date"]


def get_event_count(path):
    try:
        ec = get_event_count_quick(path)
    except BaseException:
        config = get_rtdc_config(path)
        ec = config["experiment"]["event count"]
    return ec


def get_event_count_quick(fname):
    """Get the number of events in a data set

    Parameters
    ----------
    fname: str
        Path to an experimental data file. The file format is
        determined from the file extension (tdms or rtdc).

    Returns
    -------
    event_count: int
        The number of events in the data set

    Notes
    -----
    For tdms-based data sets, there are multiple ways of determining
    the number of events, which are used in the following order
    (according to which is faster):
    1. The MX_log.ini file "Events" tag
    2. The tdms file (very slow, because it loads the entire tdms file)
       The values obtained with this method are cached on disk to
       speed up future calls with the same argument.
    """
    fname = pathlib.Path(fname).resolve()
    ext = fname.suffix

    if ext == ".rtdc":
        with h5py.File(fname, mode="r") as h5:
            event_count = h5.attrs["experiment:event count"]
    elif ext == ".tdms":
        mdir = fname.parent
        mid = fname.name.split("_")[0]
        # possible data sources
        logf = mdir / (mid + "_log.ini")
        if logf.exists():
            # 1. The MX_log.ini file "Events" tag
            with logf.open(encoding='utf-8') as fd:
                logd = fd.readlines()
            for ll in logd:
                if ll.strip().startswith("Events:"):
                    event_count = int(ll.split(":")[1])
                    break
        else:
            # 2. Open the tdms file
            with nptdms.TdmsFile.open(fname) as tdmsfd:
                event_count = len(tdmsfd["Cell Track"]["time"])
    else:
        raise ValueError("`fname` must be an .rtdc or .tdms file!")

    return event_count


def get_flow_rate(fname):
    """Get the flow rate of a data set

    Parameters
    ----------
    fname: str
        Path to an experimental data file. The file format is
        determined from the file extenssion (tdms or rtdc).

    Returns
    -------
    flow_rate: float
        The flow rate [µL/s] of the data set
    """
    fname = pathlib.Path(fname).resolve()
    ext = fname.suffix

    if ext == ".rtdc":
        with h5py.File(fname, mode="r") as h5:
            flow_rate = h5.attrs["setup:flow rate"]
    elif ext == ".tdms":
        name = fname.name
        path = fname.parent
        mx = name.split("_")[0]
        para = path / (mx + "_para.ini")
        if para.exists():
            camcfg = rt_config.load_from_file(para)
            flow_rate = camcfg["general"]["flow rate [ul/s]"]
        else:
            # analyze the filename
            warnings.warn("{}: trying to manually find flow rate.".
                          format(fname))
            flow_rate = float(fname.split("ul_s")[0].split("_")[-1])
    else:
        raise ValueError("`fname` must be an .rtdc or .tdms file!")

    return flow_rate


def get_rtdc_config(path):
    return get_rtdc_meta(path)[0]


def get_rtdc_logs(path):
    return get_rtdc_meta(path)[1]


@functools.lru_cache(maxsize=1000)
def get_rtdc_meta(path):
    with dclab.new_dataset(path) as ds:
        config = ds.config.copy()
        logs = copy.deepcopy(dict(ds.logs))
        ic = dclab.rtdc_dataset.check.IntegrityChecker(ds)
        if len(ic.check_shapein_issue3_bad_medium()):
            # https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3
            # Try to automatically determine the medium.
            swini = path.with_name(
                path.name.split("_")[0] + "_SoftwareSettings.ini")
            if swini.exists():
                lines = [ll.strip() for ll in swini.read_text().split("\n")]
                if ("Buffer_Medium_ID=0" in lines
                        and "Buffer_Medium=CellCarrierB" in lines):
                    config["setup"]["medium"] = "CellCarrier"
                    warnings.warn("Automatically set medium to 'CellCarrier', "
                                  "because of ID mismatch in '{}' (Shape-In "
                                  "issue #3).".format(swini),
                                  MetadataEditedWarning)
    return config, logs


def get_run_index(fname):
    fname = pathlib.Path(fname).resolve()
    ext = fname.suffix
    if ext == ".rtdc":
        with h5py.File(fname, mode="r") as h5:
            run_index = h5.attrs["experiment:run index"]
    elif ext == ".tdms":
        name = fname.name
        run_index = int(name.split("_")[0].strip("Mm "))
    return run_index


def get_sample_name(fname):
    fname = pathlib.Path(fname).resolve()
    ext = fname.suffix
    if ext == ".rtdc":
        with h5py.File(fname, mode="r") as h5:
            sample = h5.attrs["experiment:sample"]
    elif ext == ".tdms":
        sample = fmt_tdms.get_project_name_from_path(fname)
    if isinstance(sample, bytes):
        sample = sample.decode("utf-8")
    return sample



import shutil
import tempfile
import time

from PyQt5 import QtCore


TMPDIR = tempfile.mkdtemp(prefix=time.strftime(
    "dckit_test_%H.%M_"))

pytest_plugins = ["pytest-qt"]


def pytest_configure(config):
    """
    Allows plugins and conftest files to perform initial configuration.
    This hook is called for every plugin and initial conftest
    file after command line options have been parsed.
    """
    tempfile.tempdir = TMPDIR
    QtCore.QCoreApplication.setOrganizationName("DC-Analysis")
    QtCore.QCoreApplication.setOrganizationDomain("dc.readthedocs.io")
    QtCore.QCoreApplication.setApplicationName("DCKit")
    QtCore.QSettings.setDefaultFormat(QtCore.QSettings.IniFormat)
    settings = QtCore.QSettings()
    settings.setIniCodec("utf-8")
    settings.setValue("check for updates", 0)
    settings.sync()


def pytest_unconfigure(config):
    """
    called before test process is exited.
    """
    shutil.rmtree(TMPDIR, ignore_errors=True)



import copy
import html
import json
import functools
import numbers

import pkg_resources
import warnings

import dclab
from dclab.features.emodulus.viscosity import KNOWN_MEDIA
import numpy as np
from PyQt5 import uic, QtWidgets

from . import meta_tool
from .wait_cursor import show_wait_cursor


class IntegrityCheckDialog(QtWidgets.QDialog):
    #: A dictionary of missing metadata keys
    editable_metadata = {}
    #: Remembers user-defined metadata
    user_metadata = {}
    #: Global metadata defaults - Do not update any dictionary with
    #: this. It may override things! Instead, use the classmethod
    #: `metadata_from_path` or the function `get_metadata_value`.
    default_metadata = {}

    def __init__(self, parent, path, *args, **kwargs):
        QtWidgets.QDialog.__init__(self, parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename("dckit", "dlg_icheck.ui")
        uic.loadUi(path_ui, self)

        self.path = path
        #: metadata (remember across instances)
        if self.path not in IntegrityCheckDialog.user_metadata:
            IntegrityCheckDialog.user_metadata[self.path] = {}
        self.metadata = IntegrityCheckDialog.user_metadata[self.path]
        #: missing metadata keys
        if self.path not in IntegrityCheckDialog.editable_metadata:
            IntegrityCheckDialog.editable_metadata[self.path] = {}
        self.editables = IntegrityCheckDialog.editable_metadata[self.path]
        #: return state ("unchecked", "passed", "incomplete", "tolerable")
        self.state = "unchecked"
        # check and fill data
        self.populate_ui()
        # signals and slots
        self.toolButton_global.clicked.connect(self.on_global)

    @classmethod
    def metadata_from_path(cls, path):
        """Return the missing metadata for a specific path

        This makes use of user_metadata, editable_metadata, and
        default_metadata.

        default_metadata is only used if the data are not present
        in user_metadata and given in editable_metadata.
        """
        if path not in cls.user_metadata:
            # first create the class to populate ´editable_metadata´
            IntegrityCheckDialog(None, path)
        editables = cls.editable_metadata[path]
        userdata = cls.user_metadata[path]
        defaults = cls.default_metadata
        metadata = {}
        # first fill up the default values (if applicable)
        for sec in editables:
            if sec not in metadata:
                metadata[sec] = {}
            for key in editables[sec]:
                if sec in defaults and key in defaults[sec]:
                    metadata[sec][key] = defaults[sec][key]
        # then fill up the userdata
        for sec in userdata:
            if sec not in metadata:
                metadata[sec] = {}
            for key in userdata[sec]:
                metadata[sec][key] = userdata[sec][key]
        return metadata

    def populate_ui(self):
        """Run check and set missing UI elements"""
        # perform initial check
        cues = self.check(use_metadata=False)
        # Fill in user-changeable things
        miss_count = 0
        wrong_count = 0
        self.user_widgets = {}
        for cue in cues:
            if cue.category in ["metadata missing", "metadata wrong"]:
                # label
                sec = cue.cfg_section
                key = cue.cfg_key
                # standard choices we know of
                if sec == "setup" and key == "medium":
                    cue.cfg_choices = KNOWN_MEDIA + ["other"]
                lab = QtWidgets.QLabel("[{}]: {}".format(sec, key))
                lab.setToolTip(dclab.dfn.get_config_value_descr(sec, key))
                if cue.level == "violation":
                    lab.setStyleSheet('color: #A50000')
                else:
                    lab.setStyleSheet('color: #7A6500')
                if cue.category == "metadata missing":
                    self.gridLayout_missing.addWidget(lab, miss_count, 0)
                else:
                    self.gridLayout_wrong.addWidget(lab, wrong_count, 0)
                # control
                if cue.cfg_choices is None:
                    dt = dclab.dfn.get_config_value_type(cue.cfg_section,
                                                         cue.cfg_key)
                    if dt is str:
                        wid = QtWidgets.QLineEdit(self)
                        wid.setText(self.get_metadata_value(sec, key) or "")
                    elif dt == numbers.Integral:
                        wid = QtWidgets.QSpinBox(self)
                        wid.setMinimum(-1337)
                        wid.setMaximum(999999999)
                        value = self.get_metadata_value(sec, key) or -1337
                        wid.setValue(value)
                    elif dt == numbers.Number:
                        wid = QtWidgets.QDoubleSpinBox(self)
                        wid.setMinimum(-1337)
                        wid.setMaximum(999999999)
                        wid.setDecimals(5)
                        value = self.get_metadata_value(sec, key) or -1337
                        wid.setValue(value)
                    elif dt == (bool, np.bool_):
                        wid = QtWidgets.QComboBox(self)
                        wid.addItem("Please select", "no selection")
                        wid.addItem("True", "true")
                        wid.addItem("False", "false")
                        curdata = str(
                            self.get_metadata_value(sec, key)).lower()
                        idx = wid.findData(curdata)
                        wid.setCurrentIndex(max(0, idx))
                    else:
                        raise ValueError("No action specified '{}'".format(dt))
                else:
                    wid = QtWidgets.QComboBox(self)
                    wid.setEditable(True)  # allow user edits
                    # Set placeholder Text for lineEdit, since its editable.
                    wid.lineEdit().setPlaceholderText("Please select")
                    for item in cue.cfg_choices:
                        # Text comboboxes have no "data", because they
                        # are editable and the UI user cannot set any data.
                        wid.addItem(item)
                    wid.setCurrentText(self.get_metadata_value(sec, key))
                if cue.category == "metadata missing":
                    self.gridLayout_missing.addWidget(wid, miss_count, 1)
                    miss_count += 1
                else:
                    self.gridLayout_wrong.addWidget(wid, wrong_count, 1)
                    wrong_count += 1
                # remember all widgets for saving metadata later
                if sec not in self.user_widgets:
                    self.user_widgets[sec] = {}
                self.user_widgets[sec][key] = wid
                # remember which keys were missing
                if sec not in self.editables:
                    self.editables[sec] = []
                if key not in self.editables[sec]:
                    self.editables[sec].append(key)
        if not miss_count:
            self.groupBox_missing.hide()
        if not wrong_count:
            self.groupBox_wrong.hide()
        # Show complete log
        cues2 = self.check(use_metadata=True, expand_section=False)
        text = ""
        colors = {"info": "k",
                  "alert": "#7A6500",
                  "violation": "#A50000"}
        for cue in cues2:
            text += f"<div style='color:{colors[cue.level]}'>" \
                    + f"{html.escape(cue.msg)}</div>"

        self.textEdit.setText(text)

        # Logs
        logs = meta_tool.get_rtdc_logs(self.path)
        if logs:
            for log in logs:
                self.comboBox_logs.addItem(log, log)
        else:
            self.widget_logs.hide()
        self.comboBox_logs.currentIndexChanged.connect(self.on_logs)
        # save metadata (updates from defaults/global)
        self.save_current_metadata()

    def check(self, use_metadata=True, expand_section=True):
        if use_metadata:
            metadata_dump = json.dumps(self.metadata_from_path(self.path),
                                       sort_keys=True)
        else:
            metadata_dump = json.dumps({})
        cues = check_dataset(self.path, metadata_dump, expand_section)
        return cues

    @show_wait_cursor
    def done(self, r):
        if r:
            # save metadata
            self.save_current_metadata()
        # run check again
        cues = self.check(use_metadata=True, expand_section=False)
        levels = dclab.rtdc_dataset.check.ICue.get_level_summary(cues)
        if levels["violation"]:
            self.state = "failed"
        elif levels["alert"]:
            self.state = "tolerable"
        else:
            self.state = "passed"

        super(IntegrityCheckDialog, self).done(r)

    def get_metadata_value(self, sec, key):
        """Return the metadata value for a specific section and key

        This function gets the metadata from three sources with the
        following priority:

        1. self.metadata (previously saved by the user)
        2. dataset on disk (using meta_tool)
        3. self.default_metadata
        """
        value = None
        # Try user-defined values
        if sec in self.metadata and key in self.metadata[sec]:
            value = self.metadata[sec][key]
        # Try dataset
        if value is None:
            config = meta_tool.get_rtdc_config(self.path)
            if sec in config and key in config[sec]:
                value = config[sec][key]
        # Try default/global metadata
        if (value is None
            and sec in self.default_metadata
                and key in self.default_metadata[sec]):
            value = self.default_metadata[sec][key]
        return value

    def on_global(self):
        """Use the current metadata as default metadata"""
        # We will override self.default_metadata with this
        self.save_current_metadata()
        self.default_metadata.clear()
        self.default_metadata.update(copy.deepcopy(self.metadata))

    def on_logs(self):
        log = self.comboBox_logs.currentData()
        if log is None:
            return
        dlg = QtWidgets.QDialog()
        dlg.setWindowTitle("{}: {}".format(self.path.name, log))
        path_ui = pkg_resources.resource_filename("dckit", "dlg_log.ui")
        uic.loadUi(path_ui, dlg)
        dlg.label.setText(log)
        logs = meta_tool.get_rtdc_logs(self.path)
        text = "\n".join(logs[log])
        dlg.plainTextEdit.setPlainText(text)
        dlg.exec_()

    def save_current_metadata(self):
        for sec in self.user_widgets:
            for key in self.user_widgets[sec]:
                wid = self.user_widgets[sec][key]
                if isinstance(wid, QtWidgets.QComboBox):
                    value_a = wid.currentData()
                    if isinstance(value_a, str):
                        # for boolean combobox
                        if value_a == "true":
                            value = True
                        elif value_a == "false":
                            value = False
                        else:
                            # "no selection" / "Please select"
                            value = None
                    else:
                        # Text combobox (e.f. [setup]: medium)
                        assert value_a is None, "sanity check"
                        text_a = wid.currentText()
                        if text_a:
                            value = wid.currentText()
                        else:
                            # no text in combobox means no data
                            value = None
                elif isinstance(wid, (QtWidgets.QSpinBox,
                                      QtWidgets.QDoubleSpinBox)):
                    value = wid.value()
                else:
                    value = wid.text()
                if value is not None and value and value != -1337:
                    if sec not in self.metadata:
                        self.metadata[sec] = {}
                    self.metadata[sec][key] = value
                else:
                    # remove the value from the metadata
                    if sec in self.metadata and key in self.metadata[sec]:
                        self.metadata[sec].pop(key)


@functools.lru_cache(maxsize=1000)
def check_dataset(path, metadata_dump, expand_section):
    """Caching wrapper for integrity checks"""
    metadata = json.loads(metadata_dump)
    with warnings.catch_warnings(record=True) as ws:
        warnings.simplefilter("always")
        # ignore "ResourceWarning: unclosed file <_io.BufferedReader name=29"
        warnings.simplefilter("ignore", ResourceWarning)

        with dclab.new_dataset(path) as ds:
            ds.config.update(metadata)
            ic = dclab.rtdc_dataset.check.IntegrityChecker(ds)
            cues = ic.check(expand_section=expand_section)

            # Also check for medium "other" and offer to edit it
            for cue in cues:
                if cue.identifier == "Shape-In issue #3":
                    warnings.warn("DCKit attempted to fix '[setup]: medium'! "
                                  "(Shape-In issue #3)",
                                  meta_tool.MetadataEditedWarning)
                if (cue.category in ["metadata missing", "metadata wrong"]
                    and cue.cfg_section == "setup"
                        and cue.cfg_key == "medium"):
                    # The cue already exists.
                    break
            else:
                # The cue does not exist - add it.
                medium = ds.config.get("setup", {}).get("medium", "")
                if medium in ["other", ""]:
                    cues.append(dclab.rtdc_dataset.check.ICue(
                        msg="User might want to edit 'medium'",
                        level="alert",
                        category="metadata missing",
                        cfg_section="setup",
                        cfg_key="medium"))
        for ww in ws:
            cues.append(dclab.rtdc_dataset.check.ICue(
                msg="{}: {}".format(ww.category.__name__, ww.message),
                level="alert",
                category="warning"))
    return cues



import functools

from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import QCursor
from PyQt5.QtCore import Qt, QEventLoop


class ShowWaitCursor(object):
    def __enter__(self):
        QApplication.setOverrideCursor(QCursor(Qt.WaitCursor))
        # This overloaded function call makes sure that all events,
        # even those triggered during the function call, are processed.
        QApplication.processEvents(QEventLoop.AllEvents, 50)
        return self

    def __exit__(self, type, value, traceback):
        QApplication.restoreOverrideCursor()


def show_wait_cursor(func):
    """A decorator that starts and stops a wait cursor for a function call

    https://doc.qt.io/qt-5/qguiapplication.html#setOverrideCursor
    Every setOverrideCursor() must eventually be followed by a
    corresponding restoreOverrideCursor(), otherwise the stack
    will never be emptied.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with ShowWaitCursor():
            ret = func(*args, **kwargs)
        return ret
    return wrapper



import pathlib

import dclab
import h5py
import pytest
from PyQt5.QtWidgets import QFileDialog, QMessageBox

from dckit.main import DCKit
from dckit.dlg_icheck import IntegrityCheckDialog
from dckit.meta_tool import MetadataEditedWarning

from helper_methods import retrieve_data


def test_integrity_shapein_issue3(qtbot, monkeypatch):
    """Shape-In did not store the medium correctly

    https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3

    We don't do anything in this test. We just make sure that
    CellCarrierB is set automatically.
    """
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "2.2.2.0"
        h5.attrs["setup:medium"] = "CellCarrierB"
    h5path_m = h5path.with_name("M001_data.rtdc")
    h5path.rename(h5path_m)
    # Create SoftwareSettings.ini
    sinipath = h5path_m.with_name("M001_SoftwareSettings.ini")
    sinipath.write_text("Buffer_Medium_ID=0\nBuffer_Medium=CellCarrierB")
    # Proceed with loading
    path_out = pathlib.Path(h5path_m).parent
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([h5path_m])
    assert mw.tableWidget.rowCount() == 1, "sanity check"
    # Now edit the medium (create dialog manually)
    with pytest.warns(MetadataEditedWarning):
        dlg = IntegrityCheckDialog(mw, h5path_m)
    assert dlg.get_metadata_value("setup", "medium") == "CellCarrier"
    assert "setup" in dlg.user_widgets, "setup section must be there"
    assert "medium" in dlg.user_widgets["setup"], "medium must be there"
    # finish the dialog
    dlg.done(True)

    # 1. Get metadata from dict (just to be sure)
    dlg2 = IntegrityCheckDialog(mw, h5path_m)
    assert dlg2.get_metadata_value("setup", "medium") == "CellCarrier"

    # 2. Make sure the combobox is set correctly (just to be sure)
    wid2 = dlg2.user_widgets["setup"]["medium"]
    assert wid2.currentText() == "CellCarrier"

    # 3. Compress and check (just to be sure)
    paths_compressed, invalid = mw.on_task_compress()
    assert len(invalid) == 0
    assert len(paths_compressed) == 1
    with dclab.new_dataset(paths_compressed[0]) as ds:
        assert ds.config["setup"]["medium"] == "CellCarrier"


def test_integrity_shapein_issue3_control(qtbot, monkeypatch):
    """Shape-In did not store the medium correctly

    https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3

    With a functional Shape-In version, DCKit should not attempt anything.
    """
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "2.2.3.0"  # [sic]
        h5.attrs["setup:medium"] = "CellCarrierB"
    h5path_m = h5path.with_name("M001_data.rtdc")
    # Create SoftwareSettings.ini (should not have any effect, because
    # the Shape-In version is expected to be safe.
    sinipath = h5path_m.with_name("M001_SoftwareSettings.ini")
    sinipath.write_text("Buffer_Medium_ID=0\nBuffer_Medium=CellCarrierB")
    h5path.rename(h5path_m)
    # Proceed with loading
    path_out = pathlib.Path(h5path_m).parent
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([h5path_m])
    assert mw.tableWidget.rowCount() == 1, "sanity check"
    # Now edit the medium (create dialog manually)
    dlg = IntegrityCheckDialog(mw, h5path_m)
    assert dlg.get_metadata_value("setup", "medium") == "CellCarrierB"
    assert "setup" not in dlg.user_widgets, "setup section must not be there"
    # finish the dialog
    dlg.done(True)

    # 1. Get metadata from dict (just to be sure)
    dlg2 = IntegrityCheckDialog(mw, h5path_m)
    assert dlg2.get_metadata_value("setup", "medium") == "CellCarrierB"

    # 2. Compress and check (just to be sure)
    paths_compressed, invalid = mw.on_task_compress()
    assert len(invalid) == 0
    assert len(paths_compressed) == 1
    with dclab.new_dataset(paths_compressed[0]) as ds:
        assert ds.config["setup"]["medium"] == "CellCarrierB"


def test_integrity_shapein_issue3_control_2(qtbot, monkeypatch):
    """Shape-In did not store the medium correctly

    https://github.com/ZELLMECHANIK-DRESDEN/ShapeIn_Issues/issues/3

    This is a control test without SoftwareSettings.ini
    """
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    with h5py.File(h5path, "a") as h5:
        h5.attrs["setup:software version"] = "2.2.2.0"
        h5.attrs["setup:medium"] = "CellCarrierB"
    h5path_m = h5path.with_name("M001_data.rtdc")
    h5path.rename(h5path_m)
    # Proceed with loading
    path_out = pathlib.Path(h5path_m).parent
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([h5path_m])
    assert mw.tableWidget.rowCount() == 1, "sanity check"
    # Now edit the medium (create dialog manually)
    dlg = IntegrityCheckDialog(mw, h5path_m)
    assert dlg.get_metadata_value("setup", "medium") == "CellCarrierB"
    assert "setup" in dlg.user_widgets, "setup section must be there"
    # finish the dialog
    dlg.done(True)

    # 1. Get metadata from dict (just to be sure)
    dlg2 = IntegrityCheckDialog(mw, h5path_m)
    assert dlg2.get_metadata_value("setup", "medium") == "CellCarrierB"

    # 2. Compress and check (just to be sure)
    paths_compressed, invalid = mw.on_task_compress()
    assert len(invalid) == 0
    assert len(paths_compressed) == 1
    with dclab.new_dataset(paths_compressed[0]) as ds:
        assert ds.config["setup"]["medium"] == "CellCarrierB"


def test_integrity_with_medium(qtbot, monkeypatch):
    """This tests for a regression

    It must be possible to edit the medium and the medium has to be
    stored in the user-defined metadata.
    """
    path = retrieve_data("rtdc_data_traces_video.zip")
    path_out = pathlib.Path(path).parent
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    assert mw.tableWidget.rowCount() == 1, "sanity check"
    # Now edit the medium (create dialog manually)
    dlg = IntegrityCheckDialog(mw, path)
    assert dlg.get_metadata_value("setup", "medium") is None
    assert "setup" in dlg.user_widgets, "setup section must be there"
    assert "medium" in dlg.user_widgets["setup"], "medium must be there"
    # set medium to CellCarrier
    wid = dlg.user_widgets["setup"]["medium"]
    assert wid.currentData() is None
    wid.setCurrentText("CellCarrier")
    # finish the dialog
    dlg.done(True)

    # 1. Get metadata from dict
    dlg2 = IntegrityCheckDialog(mw, path)
    assert dlg2.get_metadata_value("setup", "medium") == "CellCarrier"

    # 2. Make sure the combobox is set correctly
    wid2 = dlg2.user_widgets["setup"]["medium"]
    assert wid2.currentText() == "CellCarrier"

    # 3. Convert and check
    paths_converted, invalid, errors = mw.on_task_tdms2rtdc()
    assert len(errors) == 0
    assert len(invalid) == 0
    assert len(paths_converted) == 1
    with dclab.new_dataset(paths_converted[0]) as ds:
        assert ds.config["setup"]["medium"] == "CellCarrier"


def test_integrity_with_medium_remove(qtbot, monkeypatch):
    """Same test as above but also test removal"""
    path = retrieve_data("rtdc_data_traces_video.zip")
    path_out = pathlib.Path(path).parent
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    # set medium to CellCarrier
    dlg = IntegrityCheckDialog(mw, path)
    wid = dlg.user_widgets["setup"]["medium"]
    wid.setCurrentText("CellCarrier")
    dlg.done(True)
    # reset the medium to nothing
    dlg2 = IntegrityCheckDialog(mw, path)
    wid2 = dlg2.user_widgets["setup"]["medium"]
    wid2.setCurrentText("")
    dlg2.done(True)

    # 1. Get metadata from dict
    dlg3 = IntegrityCheckDialog(mw, path)
    wid3 = dlg3.user_widgets["setup"]["medium"]
    assert dlg3.get_metadata_value("setup", "medium") is None

    # 2. Make sure the combobox is set correctly
    assert wid3.currentText() == ""


def test_online_contour_no_absdiff(qtbot, monkeypatch):
    """Test booleanness of the metadata combo box"""
    path = retrieve_data("rtdc_data_traces_video.zip")
    # modify the data to not have the [online_contour]: "no absdiff" keyword
    p_para = path.parent / "M1_para.ini"
    ptext = p_para.read_text().split("\n")
    # remove [online_contour]: "no absdiff"
    for ii in range(len(ptext)):
        if ptext[ii].strip() == "Diff_Method = 1":
            ptext.pop(ii)
            break
    p_para.write_text("\n".join(ptext))
    path_out = path.with_name("converted")
    path_out.mkdir()
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    # set value to True
    dlg = IntegrityCheckDialog(mw, path)
    wid = dlg.user_widgets["online_contour"]["no absdiff"]
    assert wid.currentIndex() == 0
    assert wid.currentData() == "no selection"
    idx = wid.findData("true")
    wid.setCurrentIndex(idx)
    assert wid.currentData() == "true"
    dlg.done(True)

    # 1. Get metadata from dict
    dlg2 = IntegrityCheckDialog(mw, path)
    assert dlg2.get_metadata_value("online_contour", "no absdiff")

    # 2. Make sure the combobox is set correctly
    wid2 = dlg2.user_widgets["online_contour"]["no absdiff"]
    assert wid2.currentData()

    # 3. Convert and check
    paths_converted, invalid, errors = mw.on_task_tdms2rtdc()
    assert len(errors) == 0
    assert len(invalid) == 0
    assert len(paths_converted) == 1
    with dclab.new_dataset(paths_converted[0]) as ds:
        assert ds.config["online_contour"]["no absdiff"]


def test_online_contour_no_absdiff_remove(qtbot, monkeypatch):
    """Test booleanness of the metadata combo box"""
    path = retrieve_data("rtdc_data_traces_video.zip")
    # modify the data to not have the [online_contour]: "no absdiff" keyword
    p_para = path.parent / "M1_para.ini"
    ptext = p_para.read_text().split("\n")
    # remove [online_contour]: "no absdiff"
    for ii in range(len(ptext)):
        if ptext[ii].strip() == "Diff_Method = 1":
            ptext.pop(ii)
            break
    p_para.write_text("\n".join(ptext))
    path_out = path.with_name("converted")
    path_out.mkdir()
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    # set value
    dlg = IntegrityCheckDialog(mw, path)
    wid = dlg.user_widgets["online_contour"]["no absdiff"]
    idx = wid.findData("true")
    wid.setCurrentIndex(idx)
    dlg.done(True)
    # reset value
    dlg2 = IntegrityCheckDialog(mw, path)
    wid2 = dlg2.user_widgets["online_contour"]["no absdiff"]
    idx = wid2.findData("no selection")
    wid2.setCurrentIndex(idx)
    dlg2.done(True)

    # 1. Get metadata from dict
    dlg3 = IntegrityCheckDialog(mw, path)
    assert dlg3.get_metadata_value("online_contour", "no absdiff") is None

    # 2. Make sure the combobox is set correctly
    wid3 = dlg3.user_widgets["online_contour"]["no absdiff"]
    assert wid3.currentData() == "no selection"



from dclab.rtdc_dataset.feat_temp import deregister_all
import h5py
import numpy as np
from PyQt5 import QtCore
from PyQt5.QtWidgets import QFileDialog, QDialog, QMessageBox
import pytest

from dckit.main import DCKit
from dckit.preferences import register_temporary_features

from helper_methods import retrieve_data


@pytest.fixture(autouse=True)
def run_around_tests():
    # Code that will run before your test, for example:
    pass
    # A test function will be run at this point
    yield
    # Code that will run after your test, for example:
    # Remove all temporary features
    deregister_all()
    settings = QtCore.QSettings()
    settings.remove("included features")


def test_task_compress_with_scalar_feature(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.with_name("compressed")
    path_out.mkdir()
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))

    # add a temporary feature to the dataset
    with h5py.File(path, "a") as h5:
        data = np.linspace(2, 3, len(h5["events/deform"]))
        h5["events/peter"] = data

    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])

    # Without any registered temporary features, the "peter" feature
    # will not be available.
    paths_cmp, _ = mw.on_task_compress()

    with h5py.File(paths_cmp[0], "r") as h5:
        assert "peter" not in h5["events"]

    # If we register the "peter" feature, then it should also be exported.
    settings = QtCore.QSettings()
    settings.setValue("included features", "peter")
    settings.setValue("feature scalar peter", "1")
    paths_cmp[0].unlink()
    register_temporary_features()
    mw.on_task_compress()

    with h5py.File(paths_cmp[0], "r") as h5:
        assert "peter" in h5["events"]
        assert np.all(data == h5["events/peter"])


def test_task_compress_with_non_scalar_feature(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.with_name("compressed")
    path_out.mkdir()
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))

    # add a temporary feature to the dataset
    with h5py.File(path, "a") as h5:
        data = np.zeros((len(h5["events/deform"]), 2, 5, 3))
        data[1, 0, 3, 1] = 10
        h5["events/hans"] = data

    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])

    # Without any registered temporary features, the "hans" feature
    # will not be available.
    paths_cmp, _ = mw.on_task_compress()

    with h5py.File(paths_cmp[0], "r") as h5:
        assert "hans" not in h5["events"]

    # If we register the "hans" feature, then it should also be exported.
    settings = QtCore.QSettings()
    settings.setValue("included features", "hans")
    settings.setValue("feature scalar hans", "0")
    paths_cmp[0].unlink()
    register_temporary_features()
    mw.on_task_compress()

    with h5py.File(paths_cmp[0], "r") as h5:
        assert "hans" in h5["events"]
        assert np.all(data == h5["events/hans"])



"""Test hdf5 logs history"""
import dclab

from dckit import history

from helper_methods import retrieve_data


def test_append_history():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")

    newlog = {"peter": "hans",
              "golem": 2}
    history.append_history(h5path, hdict=newlog)

    hlist = history.read_history(h5path)
    assert hlist[0] == newlog

    with dclab.new_dataset(h5path) as ds:
        assert "dckit-history" in ds.logs


def test_multiple():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")

    newlog = {"peter": "hans",
              "golem": 2}
    history.append_history(h5path, hdict=newlog)

    newlog2 = {"peter2": "hans",
               "golem": 4}
    history.append_history(h5path, hdict=newlog2)

    hlist = history.read_history(h5path)
    assert hlist[0] == newlog
    assert hlist[1] == newlog2


def test_unicode():
    h5path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")

    newlog = {"peter": "häns",
              "golem": 2}
    history.append_history(h5path, hdict=newlog)

    hlist = history.read_history(h5path)
    assert hlist[0] == newlog

    with dclab.new_dataset(h5path) as ds:
        assert "dckit-history" in ds.logs


if __name__ == "__main__":
    # Run all tests
    _loc = locals()
    for _key in sorted(list(_loc.keys())):
        if _key.startswith("test_") and hasattr(_loc[_key], "__call__"):
            _loc[_key]()



"""basic tests"""
import dclab
import h5py
import numpy as np
from PyQt5.QtWidgets import QFileDialog, QDialog, QMessageBox, QInputDialog

from dckit.main import DCKit


from helper_methods import retrieve_data


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = DCKit()
    main_window.close()


def test_list_entries(qtbot):
    mw = DCKit()
    qtbot.addWidget(mw)
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    mw.append_paths([path])
    meta = mw.get_metadata(0)
    assert meta["experiment"]["sample"] == "calibration_beads"


def test_task_compress(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.with_name("compressed")
    path_out.mkdir()
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))

    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    pouts, invalid = mw.on_task_compress()
    assert len(pouts) == 1
    assert len(invalid) == 0
    with dclab.new_dataset(pouts[0]) as ds, dclab.new_dataset(path) as ds0:
        assert len(ds) == len(ds0)
        assert len(ds.logs) != 0
        scf = list(set(ds.features_scalar + ds0.features_scalar))
        for feat in scf:
            assert feat in ds0
            assert feat in ds
            assert np.all(ds[feat] == ds0[feat])


def test_task_compress_and_repack_strip_issue19(qtbot, monkeypatch):
    """Same test as above, only tests whether repack and strip logs works"""
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.with_name("compressed")
    path_out.mkdir()
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))

    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    mw.checkBox_repack.setChecked(True)
    pouts, invalid = mw.on_task_compress()
    assert len(pouts) == 1
    assert len(invalid) == 0
    with dclab.new_dataset(pouts[0]) as ds, dclab.new_dataset(path) as ds0:
        assert len(ds) == len(ds0)
        assert len(ds.logs) == 0
        scf = list(set(ds.features_scalar + ds0.features_scalar))
        for feat in scf:
            assert feat in ds0
            assert feat in ds
            assert np.all(ds[feat] == ds0[feat])


def test_task_join(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.with_name("out.rtdc")
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QDialog.Accepted)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getSaveFileName",
                        lambda *args: (str(path_out), None))

    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path, path])
    mw.on_task_join()
    with dclab.new_dataset(path_out) as ds, dclab.new_dataset(path) as ds0:
        assert len(ds) == 2*len(ds0)


def test_task_metadata_sample(qtbot, monkeypatch):
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    mw = DCKit()
    qtbot.addWidget(mw)
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    mw.append_paths([path])
    mw.tableWidget.item(0, 3).setText("Peter Pan")
    mw.on_task_metadata()
    with dclab.new_dataset(path) as ds:
        assert ds.config["experiment"]["sample"] == "Peter Pan"


def test_task_split_trace(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_hdf5_rtfdc.zip")
    path_out = path.parent / "split"
    path_out.mkdir()
    # Monkeypatch
    monkeypatch.setattr(QDialog, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    monkeypatch.setattr(QInputDialog, "getInt",
                        lambda *args: [3, QMessageBox.Ok])
    # Initialize
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    paths_split, errors = mw.on_task_split()
    assert not errors
    assert len(paths_split) == 1

    paths_out = sorted(path_out.glob("*.rtdc"))
    assert len(paths_out) == 3

    for pp, size in zip(paths_out, [3, 3, 1]):
        with h5py.File(pp, mode="r") as h5:
            assert len(h5["events"]["trace"]["fl1_raw"]) == size, pp


def test_task_tdms2rtdc(qtbot, monkeypatch):
    path = retrieve_data("rtdc_data_traces_video.zip")
    path_out = path.with_name("converted")
    path_out.mkdir()
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    assert mw.tableWidget.rowCount() == 1
    paths_converted, invalid, errors = mw.on_task_tdms2rtdc()
    assert len(errors) == 0
    assert len(invalid) == 0
    assert len(paths_converted) == 1
    with dclab.new_dataset(paths_converted[0]) as ds:
        assert ds.config["setup"]["module composition"] == "Cell_Flow_2, Fluor"


def test_task_tdms2rtdc_bad_online_contour_no_absdiff(qtbot, monkeypatch):
    """This tests for a regression

    The [online_contour]: "no absdiff" keyword was somehow set to
    "please select" instead of ignoring it if the user did not choose
    anything.

    ValueError: could not convert string to float: 'please select'
    """
    path = retrieve_data("rtdc_data_traces_video.zip")
    # modify the data to not have the [online_contour]: "no absdiff" keyword
    p_para = path.parent / "M1_para.ini"
    ptext = p_para.read_text().split("\n")
    # remove [online_contour]: "no absdiff"
    for ii in range(len(ptext)):
        if ptext[ii].strip() == "Diff_Method = 1":
            ptext.pop(ii)
            break
    p_para.write_text("\n".join(ptext))
    path_out = path.with_name("converted")
    path_out.mkdir()
    # Monkeypatch message box to always return OK
    monkeypatch.setattr(QMessageBox, "exec_", lambda *args: QMessageBox.Ok)
    monkeypatch.setattr(QFileDialog, "getExistingDirectory",
                        lambda *args: str(path_out))
    mw = DCKit()
    qtbot.addWidget(mw)
    mw.append_paths([path])
    assert mw.tableWidget.rowCount() == 1
    paths_converted, invalid, errors = mw.on_task_tdms2rtdc()
    assert len(errors) == 0
    assert len(invalid) == 0
    assert len(paths_converted) == 1
    with dclab.new_dataset(paths_converted[0]) as ds:
        assert ds.config["setup"]["module composition"] == "Cell_Flow_2, Fluor"



import json
import os
import struct
import sys
import traceback
import urllib.request

from dclab.external.packaging import parse as parse_version
from PyQt5 import QtCore


class UpdateWorker(QtCore.QObject):
    finished = QtCore.pyqtSignal()
    data_ready = QtCore.pyqtSignal(dict)

    @QtCore.pyqtSlot(str, str)
    def processUpdate(self, version, ghrepo):
        mdict = check_release(ghrepo, version)
        if mdict["update available"]:
            self.data_ready.emit(mdict)
        self.finished.emit()


def check_for_update(version, ghrepo):
    thread = QtCore.QThread()
    obj = UpdateWorker()
    obj.moveToThread(thread)
    obj.finished.connect(thread.quit)
    thread.start()

    QtCore.QMetaObject.invokeMethod(obj, 'processUpdate',
                                    QtCore.Qt.QueuedConnection,
                                    QtCore.Q_ARG(str, version),
                                    QtCore.Q_ARG(str, ghrepo),
                                    )


def check_release(ghrepo="user/repo", version=None, timeout=20):
    """Check GitHub repository for latest release"""
    url = "https://api.github.com/repos/{}/releases/latest".format(ghrepo)
    if "GITHUB_TOKEN" in os.environ:
        hdr = {'authorization': os.environ["GITHUB_TOKEN"]}
    else:
        hdr = {}
    web = "https://github.com/{}/releases".format(ghrepo)
    errors = None  # error messages (str)
    update = False  # whether an update is available
    binary = None  # download link to binary file
    new_version = None  # string identifying new version
    try:
        req = urllib.request.Request(url, headers=hdr)
        data = urllib.request.urlopen(req, timeout=timeout).read()
    except BaseException:
        errors = traceback.format_exc()
    else:
        j = json.loads(data)

        newversion = j["tag_name"]

        if version is not None:
            new = parse_version(newversion)
            old = parse_version(version)
            if new > old:
                update = True
                new_version = newversion
                if hasattr(sys, "frozen"):
                    # determine which binary URL we need
                    if sys.platform == "win32":
                        nbit = 8 * struct.calcsize("P")
                        if nbit == 32:
                            dlid = "win_32bit_setup.exe"
                        else:
                            dlid = "win_64bit_setup.exe"
                    elif sys.platform == "darwin":
                        dlid = ".pkg"
                    else:
                        dlid = False
                    # search for binary download file
                    if dlid:
                        for a in j["assets"]:
                            if a["browser_download_url"].count(dlid):
                                binary = a["browser_download_url"]
                                break
    mdict = {"releases url": web,
             "binary url": binary,
             "version": new_version,
             "update available": update,
             "errors": errors,
             }
    return mdict



import encodings.idna  # noqa: F401
import os
import socket

import pytest
from dckit import update


with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
    try:
        s.connect(("www.python.org", 80))
        NET_AVAILABLE = True
    except socket.gaierror:
        # no internet
        NET_AVAILABLE = False


@pytest.mark.xfail(os.getenv("APPVEYOR") in ["true", "True"],
                   reason="does not always run on Appveyor")
@pytest.mark.xfail(os.getenv("CI") == "true",
                   reason="does not always run on travisCI")
@pytest.mark.skipif(not NET_AVAILABLE, reason="No network connection!")
def test_update_basic():
    mdict = update.check_release(ghrepo="DC-analysis/DCKit",
                                 version="0.10.0")
    assert mdict["errors"] is None
    assert mdict["update available"]
    mdict = update.check_release(ghrepo="DC-analysis/DCKit",
                                 version="8472.0.0")
    assert mdict["errors"] is None
    assert not mdict["update available"]



from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'Graphical toolkit for RT-DC data management'
name = 'dckit'
year = "2019"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version  # noqa: E402

setup(
    name=name,
    author=author,
    author_email='dev@craban.de',
    url='https://github.com/DC-analysis/DCKit',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="GPL v3",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["dclab[tdms]>=0.52.0",
                      "h5py>=3.9.0",
                      # https://github.com/imageio/imageio/issues/766
                      "imageio[ffmpeg]==2.10.3",
                      "nptdms>=0.27.0",
                      "numpy>=1.21",
                      "pyqt5",
                      ],
    python_requires='>=3.10, <4',
    entry_points={"gui_scripts": ['dckit = dckit.__main__:main']},
    keywords=["RT-DC", "deformability", "cytometry", "zellmechanik"],
    classifiers=['Operating System :: OS Independent',
                 'Programming Language :: Python :: 3',
                 'Intended Audience :: Science/Research',
                 ],
    platforms=['ALL']
)



"""Helper functions for displaying message boxes"""
from PyQt5.QtWidgets import QMessageBox


def error(message, info=None, details=None):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Critical)
    msg.setWindowTitle("Errors occured")
    msg.setText(message)
    if info:
        msg.setInformativeText(info)
    if details:
        msg.setDetailedText(details)
    msg.exec_()


def ignored(message, info=None, details=None):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Critical)
    msg.setWindowTitle("Ignored datasets")
    msg.setText(message)
    if info:
        msg.setInformativeText(info)
    if details:
        msg.setDetailedText(details)
    msg.exec_()


def nothing_todo(message="Nothing to do!"):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText(message)
    msg.setWindowTitle("Warning")
    msg.exec_()


def success(message, details=None):
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Information)
    msg.setWindowTitle("Success")
    msg.setText(message)
    if details:
        msg.setDetailedText(details)
    msg.exec_()



import pkg_resources

import dclab
from dclab.rtdc_dataset import feat_temp
from PyQt5 import uic, QtCore, QtWidgets

from . import dlg_icheck


class Preferences(QtWidgets.QDialog):
    """Preferences dialog to interact with QSettings"""
    instances = {}
    feature_changed = QtCore.pyqtSignal()

    def __init__(self, parent, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename("dckit", "preferences.ui")
        uic.loadUi(path_ui, self)
        self.settings = QtCore.QSettings()
        self.parent = parent

        #: configuration keys, corresponding widgets, and defaults
        self.config_pairs = [
            ["check for updates", self.general_check_for_updates, "1"],
            # comma-separated list of features
            ["included features", self.listWidget_features, ""],
        ]

        self.reload()

        # signals
        # features
        self.pushButton_add_feature.clicked.connect(self.on_feature_add)
        self.pushButton_remove_features.clicked.connect(self.on_feature_rm)
        # dialog
        self.btn_apply = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Apply)
        self.btn_apply.clicked.connect(self.on_settings_apply)
        self.btn_cancel = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Cancel)
        self.btn_ok = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        self.btn_ok.clicked.connect(self.on_settings_apply)
        self.btn_restore = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.RestoreDefaults)
        self.btn_restore.clicked.connect(self.on_settings_restore)

    @QtCore.pyqtSlot()
    def on_feature_add(self, feat=None, scalar=None):
        """User adds a new feature"""
        if feat is None:
            # open a dialog where the user can enter text
            feat, ok_pressed = QtWidgets.QInputDialog.getText(
                self,
                "Add new feature",
                "Please enter the feature name",
            )
        else:
            ok_pressed = True
        feat = feat.strip()
        if ok_pressed and feat:
            # ask the user whether this is a scalar feature
            if scalar is None:
                reply = QtWidgets.QMessageBox.question(
                    self,
                    "Register as scalar feature?",
                    f"Is '{feat}' a scalar feature (one scalar value per "
                    + "event as e.g. for deformation)?"
                )
                scalar = reply == QtWidgets.QMessageBox.Yes
            feat = feat.strip()
            item = QtWidgets.QListWidgetItem(
                f"{feat} " + ("(scalar)" if scalar else "(non-scalar)"),
            )
            self.listWidget_features.insertItem(-1, item)
            item.setData(100, feat)
            item.setData(101, scalar)

    @QtCore.pyqtSlot()
    def on_feature_rm(self):
        """User removes one or more features"""
        selected = self.listWidget_features.selectedItems()
        for row in range(self.listWidget_features.count()):
            if self.listWidget_features.item(row) in selected:
                self.listWidget_features.takeItem(row)

    def reload(self):
        """Read configuration or set default parameters"""
        for key, widget, default in self.config_pairs:
            value = self.settings.value(key, default)
            if isinstance(widget, QtWidgets.QCheckBox):
                widget.setChecked(bool(int(value)))
            elif isinstance(widget, QtWidgets.QLineEdit):
                widget.setText(value)
            elif widget is self.listWidget_features:
                # comma-separated list
                widget.clear()
                for feat in value.split(","):
                    scalar = bool(int(self.settings.value(
                        f"feature scalar {feat}", "0")))
                    self.on_feature_add(feat=feat, scalar=scalar)
            else:
                raise NotImplementedError("No rule for '{}'".format(key))
        register_temporary_features()

    @QtCore.pyqtSlot()
    def on_settings_apply(self):
        """Save current changes made in UI to settings and reload UI"""
        for key, widget, default in self.config_pairs:
            if isinstance(widget, QtWidgets.QCheckBox):
                value = int(widget.isChecked())
            elif isinstance(widget, QtWidgets.QLineEdit):
                value = widget.text().strip()
            elif widget is self.listWidget_features:
                # comma-separated list
                features = []
                for row in range(widget.count()):
                    item = widget.item(row)
                    feat = item.data(100)
                    features.append(feat)
                    # also write scalarness
                    self.settings.setValue(f"feature scalar {feat}",
                                           str(int(item.data(101))))
                value = ",".join(sorted(set(features)))
            else:
                raise NotImplementedError("No rule for '{}'".format(key))
            self.settings.setValue(key, value)

        # reload UI to give visual feedback
        self.reload()

    @QtCore.pyqtSlot()
    def on_settings_restore(self):
        self.settings.clear()
        self.reload()


def register_temporary_features():
    """Register all temporary features in the QSettings"""
    settings = QtCore.QSettings()
    fdata = settings.value("included features", "").strip()
    feat_temp.deregister_all()
    if fdata:
        for feat in fdata.split(","):
            if feat and not dclab.dfn.feature_exists(feat):
                scalar = bool(int(settings.value(
                    f"feature scalar {feat}")))
                feat_temp.register_temporary_feature(feat, is_scalar=scalar)
    # Clear the check_dataset cache otherwise any new feature will be
    # shown as unknown.
    dlg_icheck.check_dataset.cache_clear()
